@@ -1,8 +1,8 @@
 //
 // Cursors.cs
 //
-// Author:
-//   Miguel de Icaza (miguel@novell.com)
+// Contact:
+//   Moonlight List (moonlight-list@lists.ximian.com)
 //
 // Copyright 2007 Novell, Inc.
 //
@@ -26,16 +26,37 @@
 // WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 //
 
+using System;
+
 namespace System.Windows.Input {
 
-	public enum Cursors {
-	        Default,
-	        Arrow,
-	        Hand,
-	        Wait,
-	        IBeam,
-	        Stylus,
-	        Eraser,
-	        None
+	public static class Cursors {
+		public static Cursor None {
+			get { return new Cursor (CursorType.None); }
+		}
+		public static Cursor Arrow {
+			get { return new Cursor (CursorType.Arrow); }
+		}
+		public static Cursor SizeWE {
+			get { throw new NotImplementedException(); }
+		}
+		public static Cursor Eraser {
+			get { return new Cursor (CursorType.Eraser); }
+		}
+		public static Cursor Hand {
+			get { return new Cursor (CursorType.Hand); }
+		}
+		public static Cursor Wait {
+			get { return new Cursor (CursorType.Wait); }
+		}
+		public static Cursor SizeNS {
+			get { throw new NotImplementedException(); }
+		}
+		public static Cursor IBeam {
+			get { return new Cursor (CursorType.IBeam); }
+		}
+		public static Cursor Stylus {
+			get { return new Cursor (CursorType.Stylus); }
+		}
 	}
 }
