@@ -1,8 +1,8 @@
 //
 // DataTemplate.cs
 //
-// Author:
-//   Miguel de Icaza (miguel@novell.com)
+// Contact:
+//   Moonlight List (moonlight-list@lists.ximian.com)
 //
 // Copyright 2008 Novell, Inc.
 //
@@ -26,11 +26,15 @@
 // WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 //
 using System;
+using System.Security;
 
 namespace System.Windows {
 
 	public class DataTemplate : FrameworkTemplate {
 
+#if NET_2_1
+		[SecuritySafeCritical]
+#endif
 		public DependencyObject LoadContent ()
 		{
 			Console.WriteLine ("Not implemented: System.Windows.DataTemplate.LoadContent");
