@@ -13,8 +13,11 @@
 
 #include <glib.h>
 #include "dependencyobject.h"
+#include "error.h"
 #include "list.h"
 
+/* @Namespace=None */
+/* @ManagedDependencyProperties=None */
 class NameScope : public DependencyObject {
 	GHashTable *names;
 	bool temporary;
@@ -27,6 +30,7 @@ class NameScope : public DependencyObject {
 	virtual ~NameScope ();
 
  public:
+ 	/* @PropertyType=NameScope,Attached,GenerateAccessors */
 	static DependencyProperty *NameScopeProperty;
 	
 	NameScope ();
@@ -41,16 +45,12 @@ class NameScope : public DependencyObject {
 	void SetTemporary (bool flag) { temporary = flag; }
 	bool GetTemporary () { return temporary; }
 	
-	void MergeTemporaryScope (NameScope *scope);
+	void MergeTemporaryScope (NameScope *scope, MoonError *error);
 	
 	static NameScope *GetNameScope (DependencyObject *obj);
 	static void SetNameScope (DependencyObject *obj, NameScope *scope);
+
+	void Dump();
 };
 
-G_BEGIN_DECLS
-
-void namescope_init (void);
-
-G_END_DECLS
-
 #endif /* __MOON_NAMESCOPE_H__ */
