@@ -9,6 +9,7 @@
  * See the LICENSE file included with the distribution for details.
  */
 
+#include <config.h>
 #include <expat.h>
 #include <string.h>
 #include <math.h>
@@ -19,19 +20,21 @@
 #include "xaml.h"
 #include "runtime.h"
 #include "clock.h"
+#include "mediaelement.h"
+#include "debug.h"
 
 
-#define d(x)
-
-// warnings
-#define w(x)
-
 class ParserInternal {
 public:
 	XML_Parser parser;
+	gint32 bytes_read;
+	bool reparse;
+
 	ParserInternal () 
 	{
 		parser = XML_ParserCreate (NULL);
+		bytes_read = 0;
+		reparse = false;
 	}
 
 	~ParserInternal ()
@@ -66,7 +69,7 @@ PlaylistNode::~PlaylistNode ()
 
 PlaylistEntry::PlaylistEntry (MediaElement *element, Playlist *parent, Media *media)
 {
-	d(printf ("PlaylistEntry::PlaylistEntry (%p, %p, %p)\n", element, parent, media));
+	LOG_PLAYLIST ("PlaylistEntry::PlaylistEntry (%p, %p, %p)\n", element, parent, media);
 
 	this->parent = parent;
 	this->element = element;
@@ -92,7 +95,7 @@ PlaylistEntry::PlaylistEntry (MediaElement *element, Playlist *parent, Media *me
 
 PlaylistEntry::~PlaylistEntry ()
 {
-	d(printf ("PlaylistEntry::~PlaylistEntry ()\n"));
+	LOG_PLAYLIST ("PlaylistEntry::~PlaylistEntry ()\n");
 	
 	delete source_name;
 	g_free (full_source_name);
@@ -109,6 +112,17 @@ PlaylistEntry::~PlaylistEntry ()
 	g_free (info_url);
 }
 
+void
+PlaylistEntry::Dispose ()
+{
+	EventObject::Dispose ();
+	if (media) {
+		media->Dispose ();
+		media->unref ();
+		media = NULL;
+	}
+}
+
 Uri *
 PlaylistEntry::GetBase ()
 {
@@ -284,9 +298,9 @@ add_attribute (MediaAttributeCollection *attributes, const char *name, const cha
 		return;
 
 	MediaAttribute *attribute = new MediaAttribute ();
-	dependency_object_set_name (attribute, name);
-	media_attribute_set_value (attribute, g_strdup (attr));
-
+	attribute->SetValue (attr);
+	attribute->SetName (name);
+	
 	attributes->Add (attribute);
 	attribute->unref ();
 }
@@ -294,7 +308,7 @@ add_attribute (MediaAttributeCollection *attributes, const char *name, const cha
 void
 PlaylistEntry::PopulateMediaAttributes ()
 {
-	d(printf ("PlaylistEntry::PopulateMediaAttributes ()\n"));
+	LOG_PLAYLIST ("PlaylistEntry::PopulateMediaAttributes ()\n");
 
 	const char *abstract = NULL;
 	const char *author = NULL;
@@ -305,16 +319,14 @@ PlaylistEntry::PopulateMediaAttributes ()
 
 	PlaylistEntry *current = this;
 	MediaAttributeCollection *attributes;
-
-	Value *value = element->GetValue (MediaElement::AttributesProperty);
-	if (!value) {
+	
+	if (!(attributes = element->GetAttributes ())) {
 		attributes = new MediaAttributeCollection ();
-		element->SetValue (MediaElement::AttributesProperty, Value (attributes));
+		element->SetAttributes (attributes);
 	} else {
-		attributes = value->AsMediaAttributeCollection ();
 		attributes->Clear ();
 	}
-
+	
 	while (current != NULL) {
 		if (abstract == NULL)
 			abstract = current->GetAbstract ();
@@ -429,7 +441,7 @@ PlaylistEntry::GetFullSourceName ()
 void
 PlaylistEntry::Open ()
 {
-	d(printf ("PlaylistEntry::Open (), media = %p, FullSourceName = %s\n", media, GetFullSourceName ()));
+	LOG_PLAYLIST ("PlaylistEntry::Open (), media = %p, FullSourceName = %s\n", media, GetFullSourceName ());
 
 	if (media != NULL) {
 		element->SetMedia (media);
@@ -438,7 +450,7 @@ PlaylistEntry::Open ()
 
 	Downloader *dl = element->GetSurface ()->CreateDownloader ();
 	if (dl) {
-		dl->Open ("GET", GetFullSourceName ());
+		dl->Open ("GET", GetFullSourceName (), StreamingPolicy);
 		element->SetSourceInternal (dl, NULL);
 		dl->unref ();
 	}
@@ -447,7 +459,7 @@ PlaylistEntry::Open ()
 bool
 PlaylistEntry::Play ()
 {
-	d(printf ("PlaylistEntry::Play (), play_when_available: %s, media: %p, source name: %s\n", play_when_available ? "true" : "false", media, source_name ? source_name->ToString () : "NULL"));
+	LOG_PLAYLIST ("PlaylistEntry::Play (), play_when_available: %s, media: %p, source name: %s\n", play_when_available ? "true" : "false", media, source_name ? source_name->ToString () : "NULL");
 
 	if (media == NULL) {
 		play_when_available = true;
@@ -466,7 +478,7 @@ PlaylistEntry::Play ()
 bool
 PlaylistEntry::Pause ()
 {
-	d(printf ("PlaylistEntry::Pause ()\n"));
+	LOG_PLAYLIST ("PlaylistEntry::Pause ()\n");
 	
 	play_when_available = false;
 	element->GetMediaPlayer ()->Pause ();
@@ -476,7 +488,7 @@ PlaylistEntry::Pause ()
 void
 PlaylistEntry::Stop ()
 {
-	d(printf ("PlaylistEntry::Stop ()\n"));
+	LOG_PLAYLIST ("PlaylistEntry::Stop ()\n");
 
 	play_when_available = false;
 	element->GetMediaPlayer ()->Stop ();
@@ -496,7 +508,7 @@ PlaylistEntry::GetMedia ()
 void 	
 PlaylistEntry::SetMedia (Media *media)
 {
-	d(printf ("PlaylistEntry::SetMedia (%p), previous media: %p\n", media, this->media));
+	LOG_PLAYLIST ("PlaylistEntry::SetMedia (%p), previous media: %p\n", media, this->media);
 
 	if (this->media)
 		this->media->unref ();
@@ -540,15 +552,34 @@ Playlist::Playlist (MediaElement *element, Media *media)
 
 Playlist::~Playlist ()
 {
-	d(printf ("Playlist::~Playlist ()\n"));
+	LOG_PLAYLIST ("Playlist::~Playlist ()\n");
 	
 	delete entries;
 }
 
+void
+Playlist::Dispose ()
+{
+	PlaylistNode *node;
+	PlaylistEntry *entry;
+	
+	PlaylistEntry::Dispose ();
+
+	if (entries != NULL) {
+		node = (PlaylistNode *) entries->First ();
+		while (node != NULL) {
+			entry = node->GetEntry ();
+			if (entry != NULL)
+				entry->Dispose ();
+			node = (PlaylistNode *) node->next;
+		}
+	}
+}
+
 void
 Playlist::Init (MediaElement *element)
 {	
-	d(printf ("Playlist::Init (%p)\n", element));
+	LOG_PLAYLIST ("Playlist::Init (%p)\n", element);
 
 	this->element = element;
 	entries = new List ();
@@ -583,14 +614,14 @@ Playlist::Open ()
 {
 	PlaylistEntry *current_entry;
 
-	d(printf ("Playlist::Open ()\n"));
+	LOG_PLAYLIST ("Playlist::Open ()\n");
 	
 	current_node = (PlaylistNode *) entries->First ();
 
 	current_entry = GetCurrentEntry ();	
 	
 	while (current_entry && current_entry->HasDuration () && current_entry->GetDuration () == 0) {
-		d(printf ("Playlist::Open (), current entry (%s) has zero duration, skipping it.\n", current_entry->GetSourceName ()->ToString ()));
+		LOG_PLAYLIST ("Playlist::Open (), current entry (%s) has zero duration, skipping it.\n", current_entry->GetSourceName ()->ToString ());
 		current_node = (PlaylistNode *) current_node->next;
 		current_entry = GetCurrentEntry ();
 	}
@@ -598,7 +629,7 @@ Playlist::Open ()
 	if (current_entry)
 		current_entry->Open ();
 
-	d(printf ("Playlist::Open (): current node: %p, current entry: %p\n", current_entry, GetCurrentEntry ()));
+	LOG_PLAYLIST ("Playlist::Open (): current node: %p, current entry: %p\n", current_entry, GetCurrentEntry ());
 }
 
 void
@@ -606,7 +637,7 @@ Playlist::OnEntryEnded ()
 {
 	PlaylistEntry *current_entry;
 	
-	d (printf ("Playlist::OnEntryEnded () current_node: %p\n", current_node));
+	LOG_PLAYLIST ("Playlist::OnEntryEnded () current_node: %p\n", current_node);
 	
 	if (!current_node)
 		return;
@@ -619,7 +650,7 @@ Playlist::OnEntryEnded ()
 		current_entry->Play ();
 	}
 	
-	d(printf ("Playlist::OnEntryEnded () current_node: %p [Done]\n", current_node));
+	LOG_PLAYLIST ("Playlist::OnEntryEnded () current_node: %p [Done]\n", current_node);
 }
 
 bool
@@ -630,10 +661,10 @@ Playlist::Play ()
 
 	PlaylistEntry *current_entry = GetCurrentEntry ();
 
-	d(printf ("Playlist::Play (), current entry: %p\n", current_entry));
+	LOG_PLAYLIST ("Playlist::Play (), current entry: %p\n", current_entry);
 
 	while (current_entry && current_entry->HasDuration () && current_entry->GetDuration () == 0) {
-		d(printf ("Playlist::Open (), current entry (%s) has zero duration, skipping it.\n", current_entry->GetSourceName ()->ToString ()));
+		LOG_PLAYLIST ("Playlist::Open (), current entry (%s) has zero duration, skipping it.\n", current_entry->GetSourceName ()->ToString ());
 		OnEntryEnded ();
 		current_entry = GetCurrentEntry ();
 	}
@@ -649,7 +680,7 @@ Playlist::Pause ()
 {
 	PlaylistEntry *current_entry = GetCurrentEntry ();
 
-	d(printf ("Playlist::Pause ()\n"));
+	LOG_PLAYLIST ("Playlist::Pause ()\n");
 
 	if (!current_entry)
 		return false;
@@ -662,7 +693,7 @@ Playlist::Stop ()
 {
 	PlaylistEntry *current_entry = GetCurrentEntry ();
 
-	d(printf ("Playlist::Stop ()\n"));
+	LOG_PLAYLIST ("Playlist::Stop ()\n");
 
 	if (!current_entry)
 		return;
@@ -682,7 +713,7 @@ Playlist::PopulateMediaAttributes ()
 {
 	PlaylistEntry *current_entry = GetCurrentEntry ();
 
-	d(printf ("Playlist::PopulateMediaAttributes ()\n"));
+	LOG_PLAYLIST ("Playlist::PopulateMediaAttributes ()\n");
 
 	if (!current_entry)
 		return;
@@ -693,7 +724,7 @@ Playlist::PopulateMediaAttributes ()
 void
 Playlist::AddEntry (PlaylistEntry *entry)
 {
-	d(printf ("Playlist::AddEntry (%p)\n", entry));
+	LOG_PLAYLIST ("Playlist::AddEntry (%p)\n", entry);
 
 	entries->Append (new PlaylistNode (entry));
 	entry->unref ();
@@ -704,7 +735,7 @@ Playlist::ReplaceCurrentEntry (Playlist *pl)
 {
 	PlaylistEntry *current_entry = GetCurrentEntry ();
 
-	d(printf ("Playlist::ReplaceCurrentEntry (%p)\n", pl));
+	LOG_PLAYLIST ("Playlist::ReplaceCurrentEntry (%p)\n", pl);
 
 	int counter = 0;
 	PlaylistEntry *e = current_entry;
@@ -734,7 +765,7 @@ Playlist::ReplaceCurrentEntry (Playlist *pl)
 void
 Playlist::MergeWith (PlaylistEntry *entry)
 {
-	d(printf ("Playlist::MergeWith (%p)\n", entry));
+	LOG_PLAYLIST ("Playlist::MergeWith (%p)\n", entry);
 
 	SetBase (entry->GetBase () ? entry->GetBase ()->Clone () : NULL);
 	SetTitle (g_strdup (entry->GetTitle ()));
@@ -757,7 +788,19 @@ PlaylistParser::PlaylistParser (MediaElement *element, IMediaSource *source)
 {
 	this->element = element;
 	this->source = source;
+}
 
+void
+PlaylistParser::SetSource (IMediaSource *new_source)
+{
+	if (source)
+		source->unref ();
+	source = new_source;
+}
+
+void
+PlaylistParser::Setup ()
+{
 	playlist = NULL;
 	current_entry = NULL;
 	current_text = NULL;
@@ -773,7 +816,8 @@ PlaylistParser::PlaylistParser (MediaElement *element, IMediaSource *source)
 	XML_SetCharacterDataHandler (internal->parser, on_text);
 }
 
-PlaylistParser::~PlaylistParser ()
+void
+PlaylistParser::Cleanup ()
 {
 	kind_stack->Clear (true);
 	delete kind_stack;
@@ -782,10 +826,15 @@ PlaylistParser::~PlaylistParser ()
 		playlist->unref ();
 }
 
+PlaylistParser::~PlaylistParser ()
+{
+	Cleanup ();
+}
+
 static bool
 str_match (const char *candidate, const char *tag)
 {
-	return g_strcasecmp (candidate, tag) == 0;
+	return g_ascii_strcasecmp (candidate, tag) == 0;
 }
 
 void
@@ -927,6 +976,18 @@ time_value_from_str (PlaylistParser *parser, const char *str, TimeSpan *res)
 	return true;
 }
 
+static bool
+is_valid_protocol (const char *proto)
+{
+	return proto != NULL && (
+		g_ascii_strncasecmp (proto, "http", 4) == 0 || 
+		g_ascii_strncasecmp (proto, "https", 5) == 0 ||
+		g_ascii_strncasecmp (proto, "mms", 3) == 0 ||
+		g_ascii_strncasecmp (proto, "rtsp", 4) == 0 ||
+		g_ascii_strncasecmp (proto, "rstpt", 5) == 0
+	);
+}
+
 void
 PlaylistParser::OnStartElement (const char *name, const char **attrs)
 {
@@ -934,7 +995,7 @@ PlaylistParser::OnStartElement (const char *name, const char **attrs)
 	Uri *uri;
 	bool failed;
 
-	d(printf ("PlaylistParser::OnStartElement (%s, %p), kind = %d\n", name, attrs, kind));
+	LOG_PLAYLIST ("PlaylistParser::OnStartElement (%s, %p), kind = %d\n", name, attrs, kind);
 
 	g_free (current_text);
 	current_text = NULL;
@@ -990,16 +1051,10 @@ PlaylistParser::OnStartElement (const char *name, const char **attrs)
 					uri = new Uri ();
 					if (!uri->Parse (attrs [i+1], true)) {
 						failed = true;
-					} else if (uri->protocol == NULL) {
-						failed = true;
-					} else if (!strcase_equal (uri->protocol, "http") && 
-								!strcase_equal (uri->protocol, "https") && 
-								!strcase_equal (uri->protocol, "mms") &&
-								!strcase_equal (uri->protocol, "rtsp") && 
-								!strcase_equal (uri->protocol, "rstpt")) {
+					} else if (!is_valid_protocol (uri->protocol)) {
 						failed = true;
 					}
-					
+
 					if (!failed) {
 						GetCurrentContent ()->SetBase (uri);
 					} else {
@@ -1024,7 +1079,7 @@ PlaylistParser::OnStartElement (const char *name, const char **attrs)
 			if (str_match (attrs [i], "VALUE")) {
 				if (time_value_from_str (this, attrs [i+1], &ts)) {
 					if (GetCurrentEntry () != NULL && GetParentKind () != PlaylistNode::Ref) {
-						d(printf ("PlaylistParser::OnStartElement (%s, %p), found VALUE/timespan = %f s\n", name, attrs, TimeSpan_ToSecondsFloat (ts)));
+						LOG_PLAYLIST ("PlaylistParser::OnStartElement (%s, %p), found VALUE/timespan = %f s\n", name, attrs, TimeSpan_ToSecondsFloat (ts));
 						GetCurrentEntry ()->SetDuration (ts);
 					}
 				}
@@ -1118,7 +1173,7 @@ PlaylistParser::OnStartElement (const char *name, const char **attrs)
 			if (str_match (attrs [i], "VALUE")) {
 				if (time_value_from_str (this, attrs [i+1], &ts)) {
 					if (GetCurrentEntry () != NULL && GetParentKind () != PlaylistNode::Ref) {
-						d(printf ("PlaylistParser::OnStartElement (%s, %p), found VALUE/timespan = %f s\n", name, attrs, TimeSpan_ToSecondsFloat (ts)));
+						LOG_PLAYLIST ("PlaylistParser::OnStartElement (%s, %p), found VALUE/timespan = %f s\n", name, attrs, TimeSpan_ToSecondsFloat (ts));
 						GetCurrentEntry ()->SetStartTime (ts);
 					}
 				}
@@ -1159,13 +1214,20 @@ PlaylistParser::OnStartElement (const char *name, const char **attrs)
 	case PlaylistNode::Repeat:
 	case PlaylistNode::Param:
 	case PlaylistNode::Event:
-		ParsingError (new ErrorEventArgs (MediaError, 3006, "Unsupported ASX element"));
+		if (element->GetSurface ()->GetRelaxedMediaMode ()) {
+			LOG_PLAYLIST ("PlaylistParser::OnStartElement ('%s', %p): "
+				"Unsupported kind: %d, ignoring in relaxed mode\n", name, attrs, kind);
+		} else {
+			ParsingError (new ErrorEventArgs (MediaError, 3006, "Unsupported ASX element"));
+		}
 		break;
 	case PlaylistNode::Root:
 	case PlaylistNode::Unknown:
 	default:
-		d(printf ("PlaylistParser::OnStartElement ('%s', %p): Unknown kind: %d\n", name, attrs, kind));
-		ParsingError (new ErrorEventArgs (MediaError, 3004, "Invalid ASX element"));
+		LOG_PLAYLIST ("PlaylistParser::OnStartElement ('%s', %p): Unknown kind: %d\n", name, attrs, kind);
+		if (!element->GetSurface ()->GetRelaxedMediaMode ()) {
+			ParsingError (new ErrorEventArgs (MediaError, 3004, "Invalid ASX element"));
+		}
 		break;
 	}
 }
@@ -1176,7 +1238,7 @@ PlaylistParser::OnEndElement (const char *name)
 	PlaylistNode::Kind kind = GetCurrentKind ();
 	TimeSpan ts; 
 
-	d(printf ("PlaylistParser::OnEndElement (%s), GetCurrentKind (): %d, GetCurrentKind () to string: %s\n", name, kind, KindToString (kind)));
+	LOG_PLAYLIST ("PlaylistParser::OnEndElement (%s), GetCurrentKind (): %d, GetCurrentKind () to string: %s\n", name, kind, KindToString (kind));
 
 	switch (kind) {
 	case PlaylistNode::Abstract:
@@ -1256,7 +1318,13 @@ PlaylistParser::OnEndElement (const char *name)
 		}
 		break;
 	default:
-		d(printf ("PlaylistParser::OnEndElement ('%s'): Unknown kind %d.\n", name, kind));
+		if (kind == PlaylistNode::Param && element->GetSurface ()->GetRelaxedMediaMode ()) {
+			if (!AssertParentKind (PlaylistNode::Asx))
+				break;
+			break;		
+		}
+
+		LOG_PLAYLIST ("PlaylistParser::OnEndElement ('%s'): Unknown kind %d.\n", name, kind);
 		ParsingError (new ErrorEventArgs (MediaError, 3008, "ASX parse error"));
 		break;
 	}
@@ -1283,11 +1351,11 @@ PlaylistParser::OnText (const char *text, int len)
 
 #if DEBUG
 	char *p = g_strndup (text, len);
-	for (int i = 0; p [i] != NULL; i++)
+	for (int i = 0; p [i] != 0; i++)
 		if (p [i] == 10 || p [i] == 13)
 			p [i] = ' ';
 
-	d(printf ("PlaylistParser::OnText (%s, %d)\n", p, len));
+	LOG_PLAYLIST ("PlaylistParser::OnText (%s, %d)\n", p, len);
 	g_free (p);
 #endif
 
@@ -1310,7 +1378,7 @@ PlaylistParser::IsASX3 (IMediaSource *source)
 	if (!source->Peek (buffer, asx_header_length))
 		return false;
 		
-	return strncmp (asx_header, buffer, asx_header_length) == 0;
+	return g_ascii_strncasecmp (asx_header, buffer, asx_header_length) == 0;
 }
 
 bool
@@ -1323,13 +1391,35 @@ PlaylistParser::IsASX2 (IMediaSource *source)
 	if (!source->Peek (buffer, asx2_header_length))
 		return false;
 		
-	return strncmp (asx2_header, buffer, asx2_header_length) == 0;
+	return g_ascii_strncasecmp (asx2_header, buffer, asx2_header_length) == 0;
+}
+
+bool
+PlaylistParser::IsPossibleUrlList (IMediaSource *source)
+{
+	// This is a relaxed mode "format" that just covers a bunch of random
+	// awful corner cases found on the web for Moonshine/WMP compat, like
+	// 'ASF http://...' (seriously)
+
+	char buffer[20];
+	char *p;
+
+	memset (buffer, 0, sizeof (buffer));
+
+	if (!source->Peek (buffer, sizeof (buffer) - 1))
+		return false;
+	
+	// Ignore "IGNORE " from "IGNORE <proto>"
+	p = g_strstr_len (buffer, sizeof (buffer) - 8, " ");
+	p = p == NULL ? buffer : p + 1;
+
+	return is_valid_protocol (p);
 }
 
 bool
 PlaylistParser::ParseASX2 ()
 {
-#define BUFFER_SIZE 1024
+	const int BUFFER_SIZE = 1024;
 	int bytes_read;
 	char buffer[BUFFER_SIZE];
 	char *ref;
@@ -1341,27 +1431,27 @@ PlaylistParser::ParseASX2 ()
 
 	bytes_read = source->ReadSome (buffer, BUFFER_SIZE);
 	if (bytes_read < 0) {
-		w(fprintf (stderr, "Could not read asx document for parsing.\n"));
+		LOG_PLAYLIST_WARN ("Could not read asx document for parsing.\n");
 		return false;
 	}
 
 	key_file = g_key_file_new ();
 	if (!g_key_file_load_from_data (key_file, buffer, bytes_read,
 					G_KEY_FILE_NONE, NULL)) {
-		w(fprintf (stderr, "Invalid asx2 document.\n"));
+		LOG_PLAYLIST_WARN ("Invalid asx2 document.\n");
 		g_key_file_free (key_file);
 		return false;
 	}
 
 	ref = g_key_file_get_value (key_file, "Reference", "Ref1", NULL);
 	if (ref == NULL) {
-		w(fprintf (stderr, "Could not find Ref1 entry in asx2 document.\n"));
+		LOG_PLAYLIST_WARN ("Could not find Ref1 entry in asx2 document.\n");
 		g_key_file_free (key_file);
 		return false;
 	}
 
 	if (!g_str_has_prefix (ref, "http://") || !g_str_has_suffix (ref, "MSWMExt=.asf")) {
-		w(fprintf (stderr, "Could not find a valid uri within Ref1 entry in asx2 document.\n"));
+		LOG_PLAYLIST_WARN ("Could not find a valid uri within Ref1 entry in asx2 document.\n");
 		g_free (ref);
 		g_key_file_free (key_file);
 		return false;
@@ -1388,16 +1478,112 @@ PlaylistParser::ParseASX2 ()
 }
 
 bool
+PlaylistParser::TryFixError (gint8 *current_buffer, int bytes_read)
+{
+	if (XML_GetErrorCode (internal->parser) != XML_ERROR_INVALID_TOKEN)
+		return false;
+
+	int index = XML_GetErrorByteIndex (internal->parser);
+
+	if (index > bytes_read)
+		return false;
+	
+	LOG_PLAYLIST ("Attempting to fix invalid token error  %d.\n", index);
+
+	// OK, so we are going to guess that we are in an attribute here and walk back
+	// until we hit a control char that should be escaped.
+	char * escape = NULL;
+	while (index >= 0) {
+		switch (current_buffer [index]) {
+		case '&':
+			escape = g_strdup ("&amp;");
+			break;
+		case '<':
+			escape = g_strdup ("&lt;");
+			break;
+		case '>':
+			escape = g_strdup ("&gt;");
+			break;
+		case '\"':
+			break;
+		}
+		if (escape)
+			break;
+		index--;
+	}
+
+	if (!escape) {
+		LOG_PLAYLIST_WARN ("Unable to find an invalid escape character to fix in ASX: %s.\n", current_buffer);
+		g_free (escape);
+		return false;
+	}
+
+	int escape_len = strlen (escape);
+	int new_size = source->GetSize () + escape_len - 1;
+	int patched_size = internal->bytes_read + bytes_read + escape_len - 1;
+	gint8 * new_buffer = (gint8 *) g_malloc (new_size);
+
+	source->Seek (0, SEEK_SET);
+	source->ReadSome (new_buffer, internal->bytes_read);
+
+	memcpy (new_buffer + internal->bytes_read, current_buffer, index);
+	memcpy (new_buffer + internal->bytes_read + index, escape, escape_len);
+	memcpy (new_buffer + internal->bytes_read + index + escape_len, current_buffer + index + 1, bytes_read - index - 1); 
+	
+	source->Seek (internal->bytes_read + bytes_read, SEEK_SET);
+	source->ReadSome (new_buffer + patched_size, new_size - patched_size);
+
+	MemorySource *reparse_source = new MemorySource (source->GetMedia (), new_buffer, new_size);
+	SetSource (reparse_source);
+
+	internal->reparse = true;
+
+	g_free (escape);
+
+	return true;
+}
+
+bool
+PlaylistParser::IsValidPlaylist (IMediaSource *source)
+{
+	return PlaylistParser::IsASX3 (source) || 
+		PlaylistParser::IsASX2 (source) ||
+		(source->GetMedia ()->GetSurface ()->GetRelaxedMediaMode () && 
+			PlaylistParser::IsPossibleUrlList (source));
+}
+
+MediaResult
 PlaylistParser::Parse ()
 {
-	d(printf ("PlaylistParser::Parse ()\n"));
+	bool result;
+	gint64 last_available_pos;
+	gint64 size;
 
-	if (!this->IsASX3 (source) && this->IsASX2 (source)) {
-		/* Parse as a asx2 mms file */
-		return this->ParseASX2 ();
-	} else {
-		return this->ParseASX3 ();
-	}
+	LOG_PLAYLIST ("PlaylistParser::Parse ()\n");
+
+	do {
+		Setup ();
+
+		// Don't try to parse anything until we have all the data.
+		size = source->GetSize ();
+		last_available_pos = source->GetLastAvailablePosition ();
+		if (size != -1 && last_available_pos != -1 && size != last_available_pos)
+			return MEDIA_NOT_ENOUGH_DATA; 
+
+		if (!PlaylistParser::IsASX3 (source) && PlaylistParser::IsASX2 (source)) {
+			/* Parse as a asx2 mms file */
+			result = this->ParseASX2 ();
+		} else if (element->GetSurface ()->GetRelaxedMediaMode () && PlaylistParser::IsPossibleUrlList (source)) {
+			result = this->ParsePossibleUrlList ();
+		} else {
+			result = this->ParseASX3 ();
+		}
+
+		if (internal->reparse)
+			Cleanup ();
+	} while (internal->reparse);
+
+	return result ? MEDIA_SUCCESS : MEDIA_FAIL;
 }
 
 bool
@@ -1407,7 +1593,7 @@ PlaylistParser::ParseASX3 ()
 	void *buffer;
 
 // asx documents don't tend to be very big, so there's no need for a big buffer
-#define BUFFER_SIZE 1024
+	const int BUFFER_SIZE = 1024;
 
 	for (;;) {
 		buffer = XML_GetBuffer(internal->parser, BUFFER_SIZE);
@@ -1421,19 +1607,80 @@ PlaylistParser::ParseASX3 ()
 			fprintf (stderr, "Could not read asx document for parsing.\n");
 			return false;
 		}
-		
+
 		if (!XML_ParseBuffer (internal->parser, bytes_read, bytes_read == 0)) {
-			ParsingError (new ErrorEventArgs (MediaError, 3000, XML_ErrorString (XML_GetErrorCode (internal->parser))));
+			if (!TryFixError ((gint8 *) buffer, bytes_read))
+				ParsingError (new ErrorEventArgs (MediaError, 3000, g_strdup_printf ("%s  (%d, %d)", XML_ErrorString (XML_GetErrorCode (internal->parser)),
+												     (int) XML_GetCurrentLineNumber (internal->parser),
+												     (int) XML_GetCurrentColumnNumber (internal->parser))));
 			return false;
 		}
-		
+
 		if (bytes_read == 0)
 			break;
+
+		internal->bytes_read += bytes_read;
 	}
 
 	return playlist != NULL;
 }
 
+bool
+PlaylistParser::ParsePossibleUrlList ()
+{
+	const int BUFFER_SIZE = 1024;
+	int bytes_read, i;
+	char buffer[BUFFER_SIZE];
+	char **lines;
+	
+	playlist_version = 3;
+
+	bytes_read = source->ReadSome (buffer, BUFFER_SIZE);
+	if (bytes_read <= 0) {
+		LOG_PLAYLIST_WARN ("Could not read possible playlist document for parsing.\n");
+		return false;
+	}
+
+	buffer[bytes_read] = 0;
+	lines = g_strsplit_set (buffer, "\r\n", -1);
+
+	current_entry = NULL;
+	playlist = NULL;
+
+	for (i = 0; lines[i]; i++) {
+		char *p = g_strstr_len (lines[i], 12, " ");
+		p = p == NULL ? lines[i] : p + 1;
+		if (is_valid_protocol (p)) {
+			Uri *uri = new Uri ();
+			if (uri->Parse (p)) {
+				if (playlist == NULL) {
+					playlist = new Playlist (element, source);
+				}
+
+				PlaylistEntry *entry = new PlaylistEntry (element, playlist);
+				entry->SetSourceName (uri);
+				playlist->AddEntry (entry);
+				
+				if (current_entry == NULL) {
+					current_entry = entry;
+				}
+			} else {
+				LOG_PLAYLIST_WARN ("Could not parse URI from possible playlist: %s.\n", p);
+				delete uri;
+			}
+		}
+	}
+
+	g_strfreev (lines);
+
+	if (playlist == NULL || current_entry == NULL) {
+		LOG_PLAYLIST_WARN ("No valid URIs in possible playlist\n");
+		return false;
+	}
+
+	return true;
+}
+
 PlaylistEntry *
 PlaylistParser::GetCurrentContent ()
 {
@@ -1459,13 +1706,13 @@ void
 PlaylistParser::PushCurrentKind (PlaylistNode::Kind kind)
 {
 	kind_stack->Append (new KindNode (kind));
-	d(printf ("PlaylistParser::Push (%d)\n", kind));
+	LOG_PLAYLIST ("PlaylistParser::Push (%d)\n", kind);
 }
 
 void
 PlaylistParser::PopCurrentKind ()
 {
-	d(printf ("PlaylistParser::PopCurrentKind (), current: %d\n", ((KindNode *)kind_stack->Last ())->kind));
+	LOG_PLAYLIST ("PlaylistParser::PopCurrentKind (), current: %d\n", ((KindNode *)kind_stack->Last ())->kind);
 	kind_stack->Remove (kind_stack->Last ());
 }
 
@@ -1486,7 +1733,7 @@ PlaylistParser::GetParentKind ()
 bool
 PlaylistParser::AssertParentKind (int kind)
 {
-	d(printf ("PlaylistParser::AssertParentKind (%d), GetParentKind: %d, result: %d\n", kind, GetParentKind (), GetParentKind () & kind));
+	LOG_PLAYLIST ("PlaylistParser::AssertParentKind (%d), GetParentKind: %d, result: %d\n", kind, GetParentKind (), GetParentKind () & kind);
 
 	if (GetParentKind () & kind)
 		return true;
@@ -1499,7 +1746,7 @@ PlaylistParser::AssertParentKind (int kind)
 void
 PlaylistParser::ParsingError (ErrorEventArgs *args)
 {
-	d(printf ("PlaylistParser::ParsingError (%p)\n", args));
+	LOG_PLAYLIST ("PlaylistParser::ParsingError (%s)\n", args->error_message);
 	
 	XML_StopParser (internal->parser, false);
 	element->MediaFailed (args);
@@ -1542,7 +1789,7 @@ PlaylistParser::StringToKind (const char *str)
 		}
 	}
 
-	d(printf ("PlaylistParser::StringToKind ('%s') = %d\n", str, kind));
+	LOG_PLAYLIST ("PlaylistParser::StringToKind ('%s') = %d\n", str, kind);
 
 	return kind;
 }
@@ -1559,7 +1806,7 @@ PlaylistParser::KindToString (PlaylistNode::Kind kind)
 		}
 	}
 
-	d(printf ("PlaylistParser::KindToString (%d) = '%s'\n", kind, result));
+	LOG_PLAYLIST ("PlaylistParser::KindToString (%d) = '%s'\n", kind, result);
 
 	return result;
 }
